from decimal import Decimal
from typing import Tuple
from xml.etree.ElementTree import Element

from moneyed import Money


def money_to_amount_and_currency(money: Money) -> Tuple[int, str]:
    amount = int(money.amount * 100)
    currency_code = money.currency.code
    return amount, currency_code


def parse_money(e: Element) -> Money:
    return Money(
        amount=Decimal(e.find('amount').text) / 100,  # type: ignore
        currency=e.find('currency').text  # type: ignore
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-10-26 02:55
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('datatrans', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='aliasregistration',
            name='expiry_month',
            field=models.IntegerField(default=1, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(12)]),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='aliasregistration',
            name='expiry_year',
            field=models.IntegerField(default=1, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(99)]),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='payment',
            name='expiry_month',
            field=models.IntegerField(blank=True, null=True, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(12)]),
        ),
        migrations.AlterField(
            model_name='payment',
            name='expiry_year',
            field=models.IntegerField(blank=True, null=True, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(99)]),
        ),
        migrations.AlterField(
            model_name='refund',
            name='expiry_month',
            field=models.IntegerField(blank=True, null=True, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(12)]),
        ),
        migrations.AlterField(
            model_name='refund',
            name='expiry_year',
            field=models.IntegerField(blank=True, null=True, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(99)]),
        ),
    ]



"""
Example views for interactive testing paying and registering credit cards with datatrans.

You should restrict access (maybe with 'staff_member_required') if you chose to add these views to your urlconf.
"""

from django.forms import CharField, TextInput, forms
from django.http import HttpRequest, HttpResponse
from django.shortcuts import render
from djmoney.forms import MoneyField

from ..config import datatrans_js_url
from ..gateway import build_payment_parameters, build_register_credit_card_parameters
from ..models import CLIENT_REF_FIELD_SIZE


class PayForm(forms.Form):
    amount = MoneyField(min_value=0, default_currency='CHF')
    client_ref = CharField(
        required=True,
        max_length=CLIENT_REF_FIELD_SIZE,
        widget=TextInput(attrs={'size': CLIENT_REF_FIELD_SIZE})
    )


def pay(request: HttpRequest) -> HttpResponse:
    if request.method == 'POST':
        form = PayForm(request.POST)
        if form.is_valid():
            amount = form.cleaned_data['amount']
            parameters = build_payment_parameters(
                amount=amount,
                client_ref=form.cleaned_data['client_ref']
            )
            context = {
                'title': 'Pay {}'.format(amount),
                'datatrans_js_url': datatrans_js_url,
            }
            context.update(parameters._asdict())

            return render(request, 'datatrans/example/datatrans_form.html', context)
    else:
        form = PayForm()

    return render(request, 'datatrans/example/form.html', {
        'title': 'Payment',
        'form': form
    })


###################################################################

class RegisterCreditCardForm(forms.Form):
    client_ref = CharField(
        required=True,
        max_length=CLIENT_REF_FIELD_SIZE,
        widget=TextInput(attrs={'size': CLIENT_REF_FIELD_SIZE})
    )


def register_credit_card(request: HttpRequest) -> HttpResponse:
    if request.method == 'POST':
        form = RegisterCreditCardForm(request.POST)
        if form.is_valid():
            parameters = build_register_credit_card_parameters(client_ref=form.cleaned_data['client_ref'])
            context = {
                'title': 'Register credit card',
                'datatrans_js_url': datatrans_js_url,
            }
            context.update(parameters._asdict())

            return render(request, 'datatrans/example/datatrans_form.html', context)
    else:
        form = RegisterCreditCardForm()

    return render(request, 'datatrans/example/form.html', {
        'title': 'Register credit card',
        'form': form
    })



from django.http import HttpRequest, HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
import structlog

from ..gateway import handle_notification

logger = structlog.get_logger()


@require_POST
@csrf_exempt
def webhook_handler(request: HttpRequest) -> HttpResponse:
    logger.info('datatrans-webhook', body=request.body)
    handle_notification(request.body)
    return HttpResponse()  # If we were able to digest the notification (be it a success or an error), we are happy.



from .notification import handle_notification
from .payment_parameters import PaymentParameters, build_payment_parameters, build_register_credit_card_parameters
from .payment_with_alias import pay_with_alias
from .refunding import refund

__all__ = [
    'pay_with_alias', 'PaymentParameters', 'build_register_credit_card_parameters', 'build_payment_parameters',
    'handle_notification', 'refund'
]



from defusedxml.ElementTree import fromstring
from structlog import get_logger
from typing import Union

from .money_xml_converters import parse_money
from .utils import text_or_else
from ..config import sign_web
from ..models import AliasRegistration, Payment

logger = get_logger()


def handle_notification(xml: str) -> None:
    notification = parse_notification_xml(xml)
    logger.debug('processing-notification', notification=notification)
    notification.save()
    notification.send_signal()


def parse_notification_xml(xml: str) -> Union[AliasRegistration, Payment]:
    """"
    Both alias registration and payments are received here.
    We can differentiate them by looking at the use-alias user-parameter (and verifying the amount is 0).

    """
    body = fromstring(xml).find('body')
    transaction = body.find('transaction')
    _user_parameters = transaction.find('userParameters')

    def get_named_parameter(name):
        return _user_parameters.find("parameter[@name='" + name + "']")

    def success():
        return transaction.get('status') == 'success'

    def parse_success():
        # From the spec: sign2 is only returned in the success case
        computed_signature = sign_web(body.get('merchantId'), transaction.find('amount').text,
                                      transaction.find('currency').text,
                                      transaction.find('uppTransactionId').text)

        sign2 = get_named_parameter('sign2').text
        if computed_signature != sign2:
            raise ValueError('sign2 did not match computed signature')

        success = transaction.find('success')
        d = dict(
            response_code=success.find('responseCode').text,
            response_message=success.find('responseMessage').text,
            authorization_code=text_or_else(success.find('authorizationCode')),
            acquirer_authorization_code=success.find('acqAuthorizationCode').text,
        )
        return {k: v for k, v in d.items() if v is not None}

    def parse_error():
        error = transaction.find('error')
        d = dict(
            error_code=error.find('errorCode').text,
            error_message=error.find('errorMessage').text,
            error_detail=error.find('errorDetail').text)

        acquirer_error_code = get_named_parameter('acqErrorCode')
        if acquirer_error_code is not None:
            d['acquirer_error_code'] = acquirer_error_code.text

        return {k: v for k, v in d.items() if v is not None}

    def parse_common_attributes():
        d = dict(
            transaction_id=transaction.find('uppTransactionId').text,
            merchant_id=body.get('merchantId'),
            client_ref=transaction.get('refno'),
            amount=parse_money(transaction))

        payment_method = transaction.find('pmethod')
        if payment_method is not None:
            d['payment_method'] = payment_method.text

        request_type = transaction.find('reqtype')
        if request_type is not None:
            d['request_type'] = request_type.text

        credit_card_country = get_named_parameter('returnCustomerCountry')
        if credit_card_country is not None:
            d['credit_card_country'] = credit_card_country.text

        expiry_month = get_named_parameter('expm')
        if expiry_month is not None:
            d['expiry_month'] = int(expiry_month.text)

        expiry_year = get_named_parameter('expy')
        if expiry_year is not None:
            d['expiry_year'] = int(expiry_year.text)

        return d

    # End of inner helper functions, we're back inside parse_notification_xml

    use_alias_parameter = get_named_parameter('useAlias')
    if use_alias_parameter is not None and use_alias_parameter.text == 'true':
        # It's an alias registration

        d = dict(parse_common_attributes())

        masked_card_number = get_named_parameter('maskedCC')
        if masked_card_number is not None:
            d['masked_card_number'] = masked_card_number.text

        card_alias = get_named_parameter('aliasCC')
        if card_alias is not None:
            d['card_alias'] = card_alias.text

        if success():
            d['success'] = True
            d.update(parse_success())
        else:
            d['success'] = False
            d.update(parse_error())

        return AliasRegistration(**d)
    else:
        # It's a payment or a charge
        if success():
            d = dict(success=True)
            cardno = get_named_parameter('cardno')
            if cardno is not None:
                d['masked_card_number'] = cardno.text
            d.update(parse_common_attributes())
            d.update(parse_success())
            return Payment(**d)
        else:
            d = dict(success=False)
            d.update(parse_common_attributes())
            d.update(parse_error())
            return Payment(**d)



# Generated by Django 2.2.4 on 2019-08-27 04:49

from django.db import migrations
import djmoney.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('datatrans', '0005_optional_aliasregistration_transaction_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='aliasregistration',
            name='amount',
            field=djmoney.models.fields.MoneyField(decimal_places=2, default_currency='CHF', max_digits=12),
        ),
        migrations.AlterField(
            model_name='payment',
            name='amount',
            field=djmoney.models.fields.MoneyField(decimal_places=2, default_currency='CHF', max_digits=12),
        ),
        migrations.AlterField(
            model_name='refund',
            name='amount',
            field=djmoney.models.fields.MoneyField(decimal_places=2, default_currency='CHF', max_digits=12),
        ),
    ]



# Generated by Django 2.0 on 2018-01-29 07:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('datatrans', '0004_alias_registration_fields_optional'),
    ]

    operations = [
        migrations.AlterField(
            model_name='aliasregistration',
            name='transaction_id',
            field=models.CharField(blank=True, max_length=18, null=True, unique=True),
        ),
    ]



import hashlib
import hmac
import os
from typing import Any, Sequence

from django.conf import settings

web_merchant_id = settings.DATATRANS['WEB_MERCHANT_ID']
_web_hmac_key = bytearray.fromhex(settings.DATATRANS['WEB_HMAC_KEY'])
mpo_merchant_id = settings.DATATRANS['MPO_MERCHANT_ID']
_mpo_hmac_key = bytearray.fromhex(settings.DATATRANS['MPO_HMAC_KEY'])

if settings.DATATRANS.get('ENVIRONMENT') == 'PRODUCTION':
    _pay_base_url = 'https://pay.datatrans.com/'
    _api_base_url = 'https://api.datatrans.com/'
else:
    _pay_base_url = 'https://pay.sandbox.datatrans.com/'
    _api_base_url = 'https://api.sandbox.datatrans.com/'

datatrans_js_url = os.path.join(_pay_base_url, 'upp/payment/js/datatrans-2.0.0.min.js')
datatrans_authorize_url = os.path.join(_api_base_url, 'upp/jsp/XML_authorize.jsp')
datatrans_processor_url = os.path.join(_api_base_url, 'upp/jsp/XML_processor.jsp')


def sign_web(*values: Any) -> str:
    return _sign(values, _web_hmac_key)


def sign_mpo(*values: Any) -> str:
    return _sign(values, _mpo_hmac_key)


def _sign(values: Sequence[Any], key: bytearray) -> str:
    s = ''.join(map(str, values))
    return hmac.new(
        key=key,
        msg=s.encode('utf-8'),
        digestmod=hashlib.sha256
    ).hexdigest()



#!/usr/bin/env python
from setuptools import setup

import datatrans

setup(
    name='django-datatrans-gateway',
    version=datatrans.__version__,
    description='Integrate django with the datatrans payment service provider',
    long_description='',
    author='Nicholas Wolff',
    author_email='nwolff@gmail.com',
    url=datatrans.__URL__,
    download_url='https://pypi.python.org/pypi/django-datatrans-gateway',
    packages=[
        'datatrans',
        'datatrans.gateway',
        'datatrans.views',
        'datatrans.migrations',
    ],
    package_data={'datatrans': [
        'templates/admin/datatrans/*.html',
        'templates/datatrans/example/*.html',
    ]},
    install_requires=[
        'Django>=2.0',
        'django-money',
        'requests',
        'defusedxml',
        'structlog',
    ],
    license=datatrans.__licence__,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)
